

package Sequence;

#
# Standard Perl 'OOP'
#
use vars qw($AUTOLOAD @ISA);
use Exporter;
use Carp;
use strict;

use Range; # not really necessary, but if you want to use subSequence then...

#
# We're so OOP, we don't need to export. <grin>
#



#
#Inheritance Package. Sequence is a base class.
# 
#
@ISA = qw(Exporter);


my %fields = (
	      name  => undef,
	      seq => undef,
	      start => undef,
	      end => undef,
	      score => undef,
	      desc => undef, # not necessarily there
	      acc  => undef, # not necessarily here
);

sub new {
    my $ref = shift;
    my $class = ref($ref) || $ref;
    
    my $self = {
	'_permitted' => \%fields,
	%fields, };

    bless $self, $class;
    return $self;
}

#
# Autoload up my field name set/get
#
sub AUTOLOAD {
    my $self = shift;
    my $type = ref($self) || carp "Ok, $self is not an object of mine!";
    my $name = $AUTOLOAD;

    # don't propagate DESTROY messages...

    $name =~ /::DESTROY/ && return;

    $name =~ s/.*://; #get only the bit we want
    unless (exists $self->{'_permitted'}->{$name} ) {
	carp "In type $type, can't access $name - probably passed a wrong variable into Sequence  ";
    }
    if (@_) {
	return $self->{$name} = shift;
    } else {
	return $self->{$name}
    }
}

sub revcomp {
    my $self = shift;
    my $warn = shift;
    my $seq;

    if(!defined $warn) {
	$warn = 1;
    }

    
    $seq = $self->seq();

    $seq =~ tr/ATGCatgc/TACGtacg/;
    if( $seq =~ /^ATGCatgcNn/ && ($warn != 0 ) ) {
	carp("You have some non ATGC characters that you are trying to reverse complement. They will become 'N's");
    }
	
    $seq =~ s/[^ATGCatgcNn]/N/g;

    $seq = reverse $seq;

    $self->seq($seq);
}

sub has_stop {
    my $self = shift;
    my $seq;

    $seq = $self->seq();

    if( $seq =~ /\*/ ) {
	return 1;
    }

    return 0;    
}

sub subseq {
    my $self = shift;
    my $start = shift;
    my $end = shift;

    if( !defined $end ) {
	carp("Sub sequence takes 2 int arguments");
    }

    if( $end > length($self->{'seq'}) ) {
	$end = length($self->{'seq'});
    }

    return substr($self->{'seq'},$start,($end-$start));
}

sub new_from_strings {
    my $ref = shift;
    my $name = shift;
    my $seq  = shift;
    my $out;


    if( !$seq ) {
	carp("Trying to make a new Sequence without a sequence!. Nope!");
	return undef;
    }
    
    $out = new Sequence;
    bless $out, $ref; # in case this was actually a different class!
    
    $out->name($name);
    $out->seq($seq);

    return $out;
}

sub clone {
    my $ref = shift;
    my $seq = shift;
    my $out;

    $out = new Sequence;
    $out->name($seq->name());
    $out->seq($seq->seq());
    return $out;
}

sub isdna {
    my $self = shift;
    
    $self->{'seq'} =~ /[^ATGCNatgcn]/ && return 0;
    return 1;
}

sub isdna_strict {
    my $self = shift;
    
    $self->{'seq'} =~ /[^ATGCatgc]/ && return 0;
    return 1;
}

sub length {
    my $self = shift;

    return length ($self->seq());
}

sub truncate {
    my $self = shift;
    my $start = shift;
    my $end   = shift;

    $self->seq(substr($self->seq(),$start-1,$end-$start+1));

    $self->start($start);
    $self->end($end);

    return 1;
}

sub recalculate {
    my $self = shift;
    
    $self->{'end'}= $self->start() + length($self->{'seq'}) -1;
}

sub uppercase {
    my $self = shift;
    my $seq;

    $seq = $self->seq();

    $seq =~ tr/[a-z]/[A-Z]/;

    $self->seq($seq);
    
}



sub get_nse {
    my $self = shift;
    my $namesep = shift;
    my $nosep = shift;
    my ($name,$start,$end);

    if( !defined $namesep) {
	$namesep = "/";
    }
    if( !defined $nosep) {
	$nosep = "-";
    }

    $name= $self->name();
    $start= $self->start();
    $end = $self->end();

    return "$name$namesep$start$nosep$end";
}


sub read_efetch_swiss34 {
    my $self = shift;
    my $name = shift;
    my ($temp,$line);

    open(EFETCH,"efetch -f -D /nfs/disk100/pubseq/swiss34/ $name|") || carp("Could not open efetch.");


    if( !($line = <EFETCH>) ) {
	return 0;
    }

    $line =~ /^>(\S+)\s/;

    $name = $1;
    $self->name($name);

    while( <EFETCH> ) {
	s/\W//g;
	$temp .= $_;
    }
    
    close(EFETCH);

    $self->seq($temp);

    return 1;
}
    

sub read_srs {
    my $self = shift;
    my $string = shift;

    open(GETZ,"getz -d '$string' |") || carp ("Could not open getz. Nope!");

    $self->read_fasta(\*GETZ);

    close(GETZ);

    return;
}
 
sub GCG_checksum {
    my $self = shift;
    my $seq;
    my $index = 0;
    my $checksum = 0;
    my $char;


    $seq = $self->seq();
    $seq =~ tr/[a-z]/[A-Z]/;

    foreach $char ( split(/[\.\-]*/, $seq)) {
	$index++;
	$checksum += ($index * (unpack("c",$char)));
	if( $index ==  57 ) {
	    $index = 0;
	}
    }

    return ($checksum % 10000);
}

sub read_fasta_file {
    my $self = shift;
    my $file = shift;
    my $count;

    if( ! open(TEMP,$file) ) {
	carp("Could not open $file for fasta reading $!");
	return undef;
    }

    $count = $self->read_fasta(\*TEMP);

    close(TEMP);

    return $count;
}
    


sub read_fasta {
    my $self = shift;
    my $file = shift;
    my ($rs,$entry,$nameline,@seqlines,$name,@words,$seqstr);
    my $count;

    #
    # Yuk. To read fasta it is nice to change the record
    # separator. But this is a global not local to file
    # So - we have to be careful to change it back again
    # 

    $rs = $/;

    $/ = "\n>";

    
    $entry = <$file>; # reads >name desc\nseq\nseq etc
   # print STDERR "Got [$entry]\n";


    $/ = $rs;

    ($nameline,@seqlines) = split(/\n/,$entry);


    ($name,@words) = split(/\s/,$nameline);

    $seqstr = join('',@seqlines);
    $seqstr =~ s/[^A-Za-z]//g;
    $name =~ s/^>//; # first one has > on it!
    $count = length $seqstr;

    $self->name($name);
    $self->desc(join(' ',@words));
    $self->seq($seqstr);
    $self->start(1);
    $self->recalculate();

    return $count;

}


sub write_fasta {
    my $self = shift;
    my $file = shift;
    my $disname = shift;
    my $name;
    my ($seq,$count,$subseq,$length,$seqsub);
    
    if( defined $disname ) {
	if( $disname == 1 ) {
	    $name = $self->get_nse();
	} else {
	    $name = $disname;
	}
    } else {
	$name = $self->name();
    }

    $seq  = $self->seq();

    print $file ">$name\n";

    $count =0;
    $length = length $seq;
    while( ($count * 60 ) < $length ) {
	$seqsub = substr($seq,$count*60,60);
	print $file "$seqsub\n";
	$count++;
    }

}

sub write_ace {
    my $self = shift;
    my $file = shift;
    my $disname = shift;
    my $name;
    my ($seq, $length, $count, $subseq, $seqsub);
    
    if( defined $disname ) {
	if( $disname == 1 ) {
	    $name = $seq->get_nse();
	} else {
	    $name = $disname;
	}
    } else {
	$name = $self->name();
    }

    

    print $file "Sequence $name\nGenomic_canonical\n\nDNA $name\n";

    $seq  = $self->seq();

    $length = length $seq;
    while( ($count * 60 ) < $length ) {
	$seqsub = substr($seq,$count*60,60);
	print $file "$seqsub\n";
	$count++;
    }
    
}

sub blast_seq {
    my $self = shift;
    my $which = shift;
    my $blast = shift;
    my $seq = shift;
    my $options = shift;
    my $msp = shift;
    my $mspoptions = shift;
    my @result;
    my $command;
    
    if (defined $msp) {
	if (defined $mspoptions) {
	    $command = "$blast $seq $options | MSPcrunch $mspoptions";
	} else {
	    $command = "$blast $seq $options | MSPcrunch -I 90.0 -d -";
	}
    } else {
	$command = "$blast $seq $options";
    }
	
    open (BLASTN, "bsub -I -q blast_farm -R blast blast$which $command |") or carp ("Can't run blast : $!\n");

    @result = <BLASTN>;
    close BLASTN;

    return @result;
	

}

# Sequence->fromEMBLStream ( \*FH, \$line )
#
# reads in the sequence from a stream, assuming it to be embl formatted
# stops when it reaches a line that does not look like part of the sequence
#

sub fromEMBLStream {
    my $class = shift;
    my $fh = shift;            # file handle to read from
    my $lineRef = shift;       # reference to current line of text
    my $self = new Sequence;   # the new sequence object
    my $line;
    my $tag;
    my $sequence = "";

    $line = $$lineRef;
    while (1) {
#	print $line;
	$tag = substr ($line, 0, 2);
	if($tag eq 'SQ') {
	    # throw away all of the sequence information - we can make it later
	    next;
	}
	if ($tag eq '  ') {
	    $line = EMBL::TextFunc->cleanLine($line);  # clean the line
	    $line =~ tr/0-9 //d;                       # remove the numbers and spaces
	    $sequence .= $line;                        # append this to what we have found so far
#	    print "About to next\n";
	    next;
	}
	last; # found nothing that we can deal with
    } continue {
#print "In continue\n";
	$line = <$fh>;
	if(not defined $line) { warn "Sequence: End of stream"; last; }
    };
    
    $self->seq($sequence);
#    $lineGroup->sequence()->recalculate();

    $$lineRef = $line;
    return $self;
}

# $sequence->toEMBLStream(\*FH)
#
# writes the sequence object out to an embl formated stream
sub toEMBLStream {
    my $self = shift;
    my $fh = shift;
    my $sequence = $self->seq();
    my $line;
    my $pos;
    my $row;
    my $pad;

    print $fh "SQ   Sequence   ",
    length $sequence, " BP; ",            # base pairs
    $sequence =~ tr/a/a/, " A; ",         # As
    $sequence =~ tr/c/c/, " C; ",         # Cs
    $sequence =~ tr/g/g/, " G; ",         # Gs
    $sequence =~ tr/t/t/, " T; ",         # Ts
    length ($sequence) - ($sequence =~ tr/acgt/acgt/), " other;\n";  # others

    for $line ($sequence =~ m/.{1,60}/g) {                     # for each chunk up to 60 bases wide
	$pos += length $line;                                  # increase our position by that amound
	$row = join " ", ($line =~ m/.{1,10}/g);               # split into chunks of 10, and join with spaces
	$pad = " " x (80 - (length $row) - (length $pos) - 5); # calculate the trailing spaces
	print $fh "     " . $row . $pad . $pos . "\n";         # print leading spaces, sequence, padding, position
    }
    return $fh;
}

# $sequence->toSwissStream(\*FH)
#
# writes the sequence object out to an swissprot formated stream
sub toSwissStream {
    my $self = shift;
    my $fh = shift;
    my $sequence = $self->seq();
    my $line;
    my $pos;
    my $row;
    my $pad;

    print $fh sprintf("SQ   SEQUENCE   %d AA;  10000 MW;  100000E NULL1;\n",$self->length());


    for $line ($sequence =~ m/.{1,60}/g) {                     # for each chunk up to 60 bases wide
	$pos += length $line;                                  # increase our position by that amound
	$row = join " ", ($line =~ m/.{1,10}/g);               # split into chunks of 10, and join with spaces
	$pad = " " x (80 - (length $row) - (length $pos) - 5); # calculate the trailing spaces
	print $fh "     " . $row . $pad . "\n";                  # print leading spaces, sequence, padding, position
    }
    return $fh;
}

sub subSequence {
    my $self = shift;
    my $range = shift;
    my $seq = new Sequence;
    my $text;

    if ($self->start > $self->end) {
	carp "The sequence object has the start and end values backwards";
	return undef;
    }
    $seq->name($self->name());

    if($range->isInverted()) {
#	print "Got a backwards range\n";
	$seq->start($range->end);
	$seq->end($range->start);
	print "Range ", $range->start, $range->end, $range->length, "\n";
	$text = substr $self->seq, $range->end - $self->start, $range->length;
	$seq->seq($text);
	$seq->revcomp();
	$seq->recalculate();
    } else {
#	print "Got forward range\n";
	$seq->start($range->start);
	$seq->end($range->end);
	$text = substr $self->seq, $range->start - $self->start, $range->length;
	$seq->seq($text);
	$seq->recalculate();
    }

    return $seq;
}

1;
__END__

=head1 NAME

Sequence.pm - Perl Object for sequences

=head1 DESCRIPTION

B<Sequence.pm> Defines the basic sequence class.
In the future it will probably be used as a base
class for Protein cDNA and Genomic classes.. but
for the moment it is fine.

Sequence is going to be used by Align and SequenceSet
classes alot (surprise surprise).

Some tricky concepts: Sequences have 'names' 'start' and
'end'. In most cases it is assummed that name,start,end is
enough to uniquely specify a sequence. Because this is the
unique idenitifier, you will find it used all over the 
packages. The abbrevation I use for name,start,end is nse,
and it is usually written name/start-end in a simple scalar.


=head3 Example1

A simple example to get us started.

	#!/usr/local/bin/perl

	use Sequence;

        # create a new Sequence

	$seq = new Sequence;

	#
	# read in from a fasta file
	#

	$seq->read_fasta_file("myfile.fa");

	#
	# another sequence, this time read in from
	# a getz pipe
	#

	open(PIPE,"getz -d '[swissprot-id:ROA1_HUMAN]' |") || die "Could not open getz pipe. Ooops!";
	$seq2 = new Sequence;

	$seq2->read_fasta(\*PIPE); # NB notice \* syntax

	#
	# truncate the first sequence
	#

	$seq->truncate(10,400);

	

=head1 Method Calls


=head2 Creating sequences

=over 5

=item C<new>

C<Sequence::new>

Creates a new sequence. Currently only uninitialized sequences
allowed (this may change). 

=item C<new_from_strings>


This creates a new sequence from the format $name,$seq ie,

    $name = "MySequence";
    $seq  = "ATTGGGGGTGGTGTGGCCCCGTGGCCC";
    $seq = new_from_strings Sequence ($name,$seq);

This is equivalent to saying

    $seq = new Sequence;
    $seq->name($name);
    $seq->seq($seq);

	
=item C<clone>

This makes a clone of a Sequence.

    $other = clone Sequence($seq); 


=back

=head2 Accessing member variables.

There are probably many cases when you
will want to access the individual fields of sequences.
As in standard Perl parlance you can 'get' the 
variables as $obj->I<field-name>() or 'set' by $obj->I<field-name>('value')

=over5

=item C<name>

Name of the sequence

=item C<seq>

Actual sequence

=item C<start>

start residue of the sequence

=item C<end>

end residue of the sequence

=back 

=head2 Manipulation of simple attributes

=over 

=item C<length>

Gives the length of the sequence

=item C<recalculate>

make's sure that start+length(seq) = end

=item C<get_nse>

B<nse> stands for "namestartend". In many cases you want
to hash on a 'unique' key for a sequence, but the unique key
cannot simply be its name, because you can have (validly) many
subsequences of a sequence in one application. get_nse gives you
a string like B<name/start-end> familiar to Pfam users. You can
pass parameters to get_nse for specifiying the separators, so
$obj->get_nse(":","#") gets you B<name:start#end>

=item isdna

Returns 0 if non ATGCN characters are there, 1 otherwise

=item isdna_strict

Returns 0 if non ATGC characters are there, 1 otherwise


=back

=head2 Example2

Using some of the attribute methods

	#
	# Let's fiddle around with some sequences
	#

	$obj = new Sequence;
	$obj->read_fasta_file("myseq.fa");


	#
	# get the name, map lower to upper case and
	# then remove all characters past the last _
	#

	$name = $obj->name()

	$name =~ tr/[a-z]/[A-Z]/;
	$name =~ s/_[^_]*$//;

	$obj->name($name);

	$obj->write_fasta(\*STDOUT);
 

=head2 I/O

reading files, as in all my modules at the moment expect
\*<FILEHANDLE> to be passed in. For function name with _file
at the end, they simply open the filename, pass the glob in,
read the Sequence and then close the file.

=over5

=item C<read_fasta>

B<read_fasta>($fileglob) reads a fasta formatted file. It leaves
the next sequence ready to be read by the next call to B<read_fasta>.

=item C<read_fasta_file> 

For people who don't want to use their own open, opens the file, 
passes it to C<read_fasta> and then closes it

=item C<write_fasta>

B<write_fasta>($fileglob), B<write_fasta>($fileglob,$name) writes
fasta formatted entry to file. If \$name is present uses that. if
\$name == 1 then it uses get_nse (name/start-end), otherwises just
uses name.

=item C<write_ace>

B<write_ace>($fileglob), B<write_ace>($fileglob,$name) writes ace 
format to file. If \$name is present uses that. If \$name == 1 then
uses get_nse (name/start-end), otherwise just uses name.

=item C<fromEMBLStream>

This function is to allow sequences to be used in the EMBL parsing
modules

Used by the embl parser to read the sequence from an embl stream.
It will eat the Sequence line, and digest the DNA from the dna lines,
returning a Sequence object representing that DNA.  It will consume
the stream upto and including the last line of DNA.

 Sequence->fromEMBLStream(\*FH, \$firstLine);

=item C<toEMBLStream>

This function is to allow sequences to be used in the EMBL parsing
modules

Writes perfectly embl formatted sequence to a stream.  This includes
the Sequence line, and all numbering, and excludes the trailing '//'.

 $mySeq->toEMBLStream(\*FH);

=back


=item uppercase

Makes the sequence uppercase

=item blast_seq

Ask Andrew Williams!

=item revcomp

If this is a dna sequence, will reverse complement it.

Non ATGC characters go to N. Calling $seq->revcomp(0) prevents error messages
about non ATGCN characters


=item has_stop

No current documentation

=item C<subSequence>

Creates a subsequence from a sequence, representing the region
delimited by a Range object.  The range co-ordinates are assumed
to refer to the position relative to the sequence's co-ordinate system.

use:

 $subseq = $sequence->subSequence($range);

=item clone

No current documentation

=item subseq

No current documentation
